/*
 * pam_cuecat by Steven Lawrance <slawrance@technologist.com> in Sept/2000
 *
 * CueCat is probably a registered trademark of Digital Convergence, and
 * I am in no way (as far as I know) connected with Digital Convergence
 * or its partners.  So please don't sue me ;).  Besides, this module
 * does NOT contain the decoding software; instead, this module relies
 * upon the CueCat Linux kernel driver, who already got a "letter of doom."
 *
 * The name "CueCat" is only being used here to refer to the device
 * that this PAM module can be used with and is not meant in any way
 * as a misappropriation of trademark.  In other words, this PAM module
 * is NOT a CueCat.
 *
 * This code is licensed under the GNU General Public License.
 *
 * This module reads codes from /dev/scanners/cuecat, looks up the
 * username:barcode_type:barcode line in /etc/cuecatpasswd, and returns
 * whether or not this user has been authenticated.
 *
 * Parameters:
 *   debug        Debugs the CueCat PAM module
 *   nowarn       Disables warning messages
 *   noerr        Disables error messages
 *   noprompt     Disables the scan prompt
 *   onerr        Either "succeed" or "fail" on an error; default=error
 *   sense        Either "allow" or "deny" access on a match; default=allow
 *   passwd       CueCat Passwd file, if not /etc/cuecatpasswd
 *   cuecat       CueCat device, if not /dev/scanners/cuecat
 *   timeout      Wait x seconds for CueCat input; default=10, 0=infinite (!)
 *   sync         Disables asynchronous reads, which disables buffer clearing
 *                and the timeouts while allowing this module to work with
 *                sync-only CueCat kernel patches.  This option is now the
 *                default in version 0.04 and was the default in 0.01 and 0.02.
 *   async        Enables asynchronous reads, allowing buffer clearing as
 *                a security measure while enabling timeouts.  This
 *                requires a CueCat kernel patch that supports the cuecat_poll
 *                function.  If async is used on a non-async CueCat kernel
 *                patch, the driver locks up.  Async was used as the only
 *                option in 0.03, but now must be explicitly enabled.
 *
 * In /etc/cuecatpasswd, each line specifies a username:barcode_type:barcode
 * and thus a user can have one or more barcodes for a match to occur by
 * having them on separate lines.
 *
 * This code began life as the pam_listfile module and was cleaned up a bit
 * in version 0.04.
 */

#include <features.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

#ifdef DEBUG
#include <assert.h>
#endif

/*
 * Here, we make a definition for the externally accessible function
 * in this file (this definition is required for static a module
 * but strongly encouraged generally) it is used to instruct the
 * modules include file to define the function prototypes.
 */

#define PAM_SM_AUTH

#include <security/pam_modules.h>
#include <security/_pam_macros.h>

/* The prefix to use on our log messages */
#define LOCAL_LOG_PREFIX "PAM-cuecat: "

/* Write an entry to PAM's log file, usually /var/log/messages */
static void _pam_log(int err, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    vsyslog(LOG_AUTH | err, format, args);
    va_end(args);
}


/* PAM calls this to see if we should let the user in */
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
    /* Internally-used variables */
    FILE *passwd, *cuecat;
    struct stat fileinfo;
    fd_set set;
    char aline[256];
    char loginline[256];
    struct timeval set_timeout;
    unsigned long len;
    char *cmpval;
    int retval;

    /* Information from PAM */
    const char *username;
    struct pam_conv *conversation;
    struct pam_message message;
    struct pam_message *pmessage = &message;
    struct pam_response *resp = NULL;
        
    /* Sysadmin-tunable parameters */
    char *passwd_file = NULL, *cuecat_device = NULL;
    int i, onerr = PAM_SERVICE_ERR, sense = 0, timeout = 10;
    char debug = 0, prompt = 1, warn = 1, errors = 1;
    char async = 0;

    D(("called."));

    /* Parse the parameters */
    for(i=0; i < argc; i++) {
        char *param, *value;
        param = (char *) malloc(strlen(argv[i])+1);
        if (param == NULL) {
            return PAM_BUF_ERR;
        }
        strcpy(param, argv[i]);
        strtok(param, "=");
        value = strtok(NULL, "=");
        if(!strcmp(param, "onerr")) {
            if(!strcmp(value, "succeed"))
                onerr = PAM_SUCCESS;
            else if(!strcmp(value, "fail"))
                onerr = PAM_SERVICE_ERR;
            else {
                _pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Bad %s configuration value: %s", param, value);
                free(param);
                return PAM_SERVICE_ERR;
            }
        } else if(!strcmp(param, "sense")) {
            if(!strcmp(value, "allow"))
                sense=0;
            else if(!strcmp(value, "deny"))
                sense=1;
            else {
                _pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Bad %s configuration value: %s", param, value);
                free(param);
                return onerr;
            }
        } else if(!strcmp(param, "debug")) {
            debug = 1;
        } else if(!strcmp(param, "nowarn")) {
            warn = 0;
        } else if(!strcmp(param, "noerr")) {
            errors = 0;
        } else if(!strcmp(param, "timeout")) {
            timeout = atol(value);
        } else if(!strcmp(param, "noprompt")) {
            prompt = 0;
        } else if(!strcmp(param, "passwd")) {
            passwd_file = (char *) malloc(strlen(value)+1);
            strcpy(passwd_file, value);
        } else if(!strcmp(param, "cuecat")) {
            cuecat_device = (char *) malloc(strlen(value)+1);
            strcpy(cuecat_device, value);
        } else if(!strcmp(param, "async")) {
            async = 1;
        } else if(!strcmp(param, "sync")) {
            async = 0;
        } else {
            _pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Unknown option: %s", param);
            free(param);
            return onerr;
        }
        free(param);
    }

    /* Define the default file locations if none were passed to us */
    if(!passwd_file) {
        passwd_file = strdup("/etc/cuecatpasswd");
    }
    if(!cuecat_device) {
        cuecat_device = strdup("/dev/scanners/cuecat");
    }

    /* Get the username from PAM */
    if(pam_get_user(pamh, &username, NULL) != PAM_SUCCESS) {
        if (errors)
            _pam_log(LOG_ERR, LOCAL_LOG_PREFIX "Call to pam_get_user() failed");
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }
    if(!username || (strlen(username) <= 0))
    {
        /* No username exists!  This is quite weird, so return an error */
        if (errors)
            _pam_log(LOG_ERR, LOCAL_LOG_PREFIX "PAM returned an empty username");
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }

    /* Write out a debug message now */
    if (debug) {
        _pam_log(LOG_INFO, LOCAL_LOG_PREFIX "Got file = %s, cuecat = %s, user = %s, sense = %d",
                 passwd_file, cuecat_device, username, sense);
    }

    /* Ensure that our cuecatpasswd file exists */
    if(warn && lstat(passwd_file,&fileinfo)) {
        if (errors)
            _pam_log(LOG_ALERT, LOCAL_LOG_PREFIX "Couldn't stat %s", passwd_file);
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }

    /* If cuecatpasswd is group/other readable/writable/executable or is not a
       normal file, return an error */
    if(warn && ((fileinfo.st_mode & S_IWOTH) || (fileinfo.st_mode & S_IROTH) ||
       (fileinfo.st_mode & S_IXOTH) || (fileinfo.st_mode & S_IWGRP) ||
       (fileinfo.st_mode & S_IRGRP) || (fileinfo.st_mode & S_IXGRP)
       || !S_ISREG(fileinfo.st_mode)))
    {
        if (errors)
            _pam_log(LOG_ALERT, LOCAL_LOG_PREFIX "%s either does not have '600' permissions or not a normal file", 
                     passwd_file);
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }

    /* Open our cuecatpasswd file */
    if((passwd = fopen(passwd_file,"r")) == NULL) {
        if (errors)
            _pam_log(LOG_ERR, LOCAL_LOG_PREFIX  "Error opening %s", passwd_file);
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }

    /* Open the CueCat (no, not literally ;) */
    if((cuecat = fopen(cuecat_device, "r")) == NULL) {
        if (errors)
            _pam_log(LOG_ERR, LOCAL_LOG_PREFIX  "Error opening CueCat on %s", cuecat_device);
        fclose(passwd);
        free(passwd_file);
        free(cuecat_device);
        return onerr;
    }

    /* If any lingering scans exist, clean them up if async */
    if (async) {
        fflush(cuecat);
        FD_ZERO(&set);
        FD_SET(fileno(cuecat), &set);
        set_timeout.tv_sec = 0;
        set_timeout.tv_usec = 0;
        while (select(FD_SETSIZE, &set, NULL, NULL, &set_timeout) && FD_ISSET(fileno(cuecat), &set))
            fgets(aline, 255, cuecat);
    }

    /* Ask the user to scan a barcode */
    if (prompt) {
        message.msg = "Scan your barcode now";
        message.msg_style = PAM_TEXT_INFO;
        pam_get_item(pamh, PAM_CONV, (const void **)&conversation);
        conversation->conv(1, (const struct pam_message **)&pmessage,
                           &resp, conversation->appdata_ptr);
        if (resp)
            _pam_drop_reply(resp, 1);
    }

    /* Set ourselves up to wait for CueCat input */
    FD_ZERO(&set);
    FD_SET(fileno(cuecat), &set);
    set_timeout.tv_sec = timeout;
    set_timeout.tv_usec = 0;

    /* See if we read a barcode or give up */
    if (!async || !timeout || select(FD_SETSIZE, &set, NULL, NULL, &set_timeout))
    {
        /* Read a brand-new, genuine, authentic CueCat input line.  MEOW :)! */
        if (!fgets(aline, 255, cuecat)) {
            if (errors)
                _pam_log(LOG_ERR,LOCAL_LOG_PREFIX "Error reading from CueCat on %s", cuecat_device);
            fclose(passwd);
            fclose(cuecat);
            free(passwd_file);
            free(cuecat_device);
            return onerr;
        }
    } else
    {
        /* Bzzt!  Too late!  We're sorry :( */
        const char *service;
        if (prompt) {
            message.msg = "Barcode Scan Timeout";
            message.msg_style = PAM_ERROR_MSG;
            pam_get_item(pamh, PAM_CONV, (const void **)&conversation);
            conversation->conv(1, (const struct pam_message **)&pmessage,
                               &resp, conversation->appdata_ptr);
            if (resp)
                _pam_drop_reply(resp, 1);
        }

        /* Return a failure */
        pam_get_item(pamh, PAM_SERVICE, (const void **)&service);
        if (warn) _pam_log(LOG_ALERT,LOCAL_LOG_PREFIX "Login timed out for %s on service %s",
                           username, service);
        fclose(passwd);
        fclose(cuecat);
        free(passwd_file);
        free(cuecat_device);
        return PAM_AUTH_ERR;
    }
    fclose(cuecat);

    /* Formulate our cuecatpasswd comparison string from the CueCat's output */
    strncpy(loginline, username, 240);
    strcat(loginline, ":");
    len = strlen(username);
    if ((cmpval = strtok(aline, ","))) {
        if ((cmpval = strtok(NULL, ","))) {
            memcpy(&(loginline[len+1]), cmpval, 3);
            strcat(loginline, ":");
            if ((cmpval = strtok(NULL, ","))) {
                unsigned long cvlen = strlen(cmpval);
                if (cmpval[cvlen - 1] <= 13)
                    cmpval[cvlen - 1] = 0; /* for CR and/or NL */
                strncpy(&(loginline[len+5]), cmpval, 255-len-3);
            }
        }
    }

    /* Write the comparison string for debug purposes */
    if (debug) {
        _pam_log(LOG_INFO, LOCAL_LOG_PREFIX "Compare string is '%s'", loginline);
    }

    /* Find a match in the cuecatpasswd file */
    retval = 1;
    while((fgets(aline, 255, passwd) != NULL) && retval) {
        if(aline[strlen(aline) - 1] == '\n')
            aline[strlen(aline) - 1] = '\0';
        retval = strcmp(aline, loginline);
    }
    fclose(passwd);
    free(passwd_file);
    free(cuecat_device);

    /* Determine whether we should call this a success or a failure */
    if ((sense && retval) || (!sense && !retval)) {
        if (debug) {
            _pam_log(LOG_INFO, LOCAL_LOG_PREFIX
                     "Returning PAM_SUCCESS, retval = %d", retval);
        }
        return PAM_SUCCESS;
    } else {
        const char *service;
        if (debug) {
            _pam_log(LOG_INFO,LOCAL_LOG_PREFIX
                     "Returning PAM_AUTH_ERR, retval = %d", retval);
        }
        pam_get_item(pamh, PAM_SERVICE, (const void **)&service);
        if (warn) _pam_log(LOG_ALERT,LOCAL_LOG_PREFIX "Refused user %s for service %s",
                           username, service);
        return PAM_AUTH_ERR;
    }
}

/* PAM calls this to set the credentials; we don't do anything here */
PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
    return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_cuecat_modstruct = {
    "pam_cuecat",
    pam_sm_authenticate,
    pam_sm_setcred,
    NULL,
    NULL,
    NULL,
    NULL,
};

#endif

/* end of module definition */

